/*
//		Borland C++ 3.0 MS-DOS Int 21h Emulation, Version 1.00
//		Copyright (C) 1991 Paradigm Systems.  All rights reserved.
//
//		Function		
//		========
//		This file contains the code for serving common MS-DOS requests.
//		The two most common services are support for dynamic memory management
//		and stream I/O functions.  While the memory management routines
//		malloc(), free(), calloc() and realloc() are self-contained, the
//		stream I/O functions are system dependent and require that serial
//		I/O drivers be written.
*/

#include	<dos.h>							/* I/O port manipulations */

#include	"typedefs.h"					/* General purpose type declarations */
#include	"dosemu.h"						/* Int 21h function/data prototypes */


#pragma	warn	-par						/* Turn off the unused parameter warning */

/*
//		Function:	Int21Handler	
//
//		This interrupt handler services any Int 21h calls from the application.
//		Since every application will have different requirements, the scope of this
//		hanlder will vary.  As implemented here, sufficient functionality exists
//		for supporting floating point arithmetic, dynamic memory management and
//		console I/O.
*/

void	interrupt	Int21Handler(
				WORD		bp,				/* Registers as saved on the stack */
				WORD		di,
				WORD		si,
				WORD		ds,
				WORD		es,
				WORD		dx,
				WORD		cx,
				WORD		bx,
				WORD		ax,
				WORD		ip,
				WORD		cs,
				WORD		psw
		)
{
	WORD		i ;							/* General purpose counter */
	BYTE_PTR	cp ;							/* General purpose BYTE pointer */
	WORD_PTR	wp ;							/* General purpose WORD pointer */
	int		c ;							/* Temporary character storage */

	MCBP		ma ;							/* Pointer to the current MCB */
	MCBP		nma ;							/* Pointer to the next MCB */
	MCBP		tmp ;							/* A working pointer to an MCB */
	WORD		mcb ;							/* Application segment identifier */


	/* Clear the carry to indicate success */
	enable() ;
	psw &= 0xfffe ;

	/* The contents of the AH register is used to determine the service */
	switch (highbyte(ax))   {
		case DOS_SETVECTOR:
			wp = (WORD_PTR) MK_FP(0, lowbyte(ax) << 2) ;
			disable() ;
			*wp++ = dx ;
			*wp = ds ;
			enable() ;
			break ;
		
		case DOS_VERSION:
			ax = 0x0a02 ;					/* Simulate MS-DOS 2.10 */
			break ;

		case DOS_GETVECTOR:
			wp = (WORD_PTR) MK_FP(0, lowbyte(ax) << 2) ;
			disable() ;
			bx = *wp++ ;
			es = *wp ;
			enable() ;
			break ;

		case DOS_READ:
			cp = (BYTE_PTR) MK_FP(ds, dx) ;
			for (i = 0; i < cx; )   {
				c = _getch() ;
				/* Check for and process backspace characters */
				if (c == '\b' && i > 0)   {
					_putch(c) ;
					_putch(' ') ;
					_putch(c) ;
					cp-- ;
					i-- ;
					continue ;
				}
				/* Check for the end of the line */
				if (c == '\r')   {
					*cp++ = _putch('\r') ;
					*cp = _putch('\n') ;
					i += 2 ;
					break ;
				}
				/* Put the character in the buffer */
				_putch(*cp++ = c) ;
				i++ ;
			}
			/* Return the number of characters read in AX */
			ax = i ;
			break ;

		case DOS_WRITE:
			cp = (BYTE_PTR) MK_FP(ds, dx) ;
			for (i = 0; i < cx; i++)   {
				/* Check for and handle a backspace charaacter */
				if (*cp == '\b')   {
					_putch(*cp) ;
					_putch(' ') ;
				}
				_putch(*cp++) ;
			}
			/* Return the number of characters written in AX */
			ax = i ;
			break ;

		case DOS_IOCTL:
			if (lowbyte(ax) == 0)   {
				dx = 0x80d3 ;
			}
			else   {
				psw |= 1 ;					/* Signal an error */
				ax = 6 ;						/* Invalid or not open handle */
			}
			break ;

/*
//		The basic scheme is to scan the allocation chain looking for the
//		specified block.  Once the block is found, shrink it (always possible)
//		or expand it as requested.  This code is used only in the compact/large
//		memory models, or when the far heap functions are used in the
//		small/medium memory models.
*/
		case DOS_REALLOC:
			/* Make a pointer to the MCB */
			ma = (MCBP) MK_FP(es - 1, 0) ;

			/* Verify we are pointing to an MCB */
			if (ma->id != 'M' && ma->id != 'Z')   {
				/* Bad block identifier - return an error code */
				ax = 9 ;
				psw |= 1 ;
				return ;
			}

			/* Now either shrink or expand it */
			if (bx > ma->size)   {
 				/* Expand - is the next block is free/sufficiently large */
				nma = (MCBP) MK_FP(es + ma->size, 0) ;
				tmp = (MCBP) MK_FP(es + bx, 0) ;
				if (nma->owner != 0 || (ma->size + nma->size) < bx)   {
					/* Return the size block that can be allocated */
					ax = 8 ;
					bx = ma->size + ((nma->owner == 0) ? nma->size : 0) ;
					psw |= 1 ;
					return ;
				}

				/* Steal the space from the second MCB */
				mcb = bx - ma->size ;
				ma->size = bx ;

				/* Initialize the new smaller MCB */
				tmp->id = nma->id ;
				tmp->owner = nma->owner ;
				tmp->size = nma->size - mcb ;
			}
			else if (bx < ma->size)   {
				/* Release - create a new MCB holding the unallocated memory */
				tmp = (MCBP) MK_FP(es + ma->size, 0) ;
				nma = (MCBP) MK_FP(es + bx, 0) ;
				nma->size = ma->size - bx - 1 ;
				nma->owner = 0 ;
				nma->id = ma->id ;

				/* Shrink the size to the specified number */
				ma->size = bx ;

				/* Combine adjacent blocks if possible */
				if (ma->id != 'Z' && tmp->owner == 0)   {
					nma->size += tmp->size + 1 ;
					nma->id = tmp->id ;
				}

				/* Insert into the chain */
				if (ma->id == 'Z')
					ma->id = 'M' ;
			}
			break ;

		default:		 						/* All other functions */
			psw |= 1 ;						/* Set the carry flag */
			break ;
	}

	/* Return with the carry flag indicating the state of the result */
	return ;
}


/*
//		Function:	InitInt21h
//
//		This function is responsible for initializing the Paradigm Systems
//		MS-DOS Int 21h emulation package.
//
//		Specific initializations include:
//			DOS emulation interrupt handlers
//			Serial console
//			Other miscellaneous initializations
*/

void	far	InitInt21h(void)
{
	/* Install the interrupt handlers */
	setvect(0x21, Int21Handler) ;

	/* Initialize the console hardware */
	InitConsole() ;

	/* Put any other initializations here */
	return ;
}
